/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef GENERATOR_DATA_SOURCE_HH
#define GENERATOR_DATA_SOURCE_HH

#include "TSeries.hh"
#include "GenParam.hh"
#include <string>
#include <map>
#include <iosfwd>
#include <vector>

namespace generator {

  /**  DataSource is a base class for DMT generator package data sources.
    *  The datasource class includes an API for both continuous and discrete
    *  data sources. It also automatically records events for later inclusion
    *  in output data.
    *  @memo Data source base class.
    *  @version 1.1; last modified November 3, 2004
    *  @author John Zweizig
    */
  class DataSource {
  public:

    /**  Data event repetition mode.
      */
    enum RepMode {
      /**  Continuous data source
       */
      kContinuous,

      /**  Generate a single event
       */
      kSingle,

      /**  Generate events with poisson statistics.
       */
      kRate,

      /**  Generate events at fixed intervals.
       */
      kFixed
    };

    typedef std::map<std::string, GenParam> ParamMap;
    typedef ParamMap::iterator param_iter;
    typedef ParamMap::const_iterator const_param_iter;

    /**  The SrcEvent class holds the generation parameters of each 
      *  generated event. It is patterned after the FrSimEvent structure 
      *  in the IGWD frame specification.
      *  @memo Data source event parameter storage class.
      */
    class SrcEvent {
    public:
      typedef std::pair<std::string, float> ParamEntry;
      typedef std::vector<ParamEntry> ParamList;
      typedef ParamList::iterator param_iter;
      typedef ParamList::const_iterator const_param_iter;

    public:
      /**  Default SrcEvent constructor.
        */
      SrcEvent(void);

      /** SrcEvent data constructor.
        */
      SrcEvent(const std::string& name, const std::string& comment, 
	       const Time& t0, Interval before, Interval after, 
	       const ParamMap& map);

      /** Get the comment field
        */
      const char* getComment(void) const;

      /**  Get the input data.
        */
      const char* getInputs(void)  const;

      /**  Get the data source name.
        */
      const char* getName(void)    const;

      /** Get the GPS time of the maximum value.
        */
      const Time& getGPSMax(void)  const;

      /**  Get the duration of the generated waveform before the maximum time.
       */
      Interval    getTimeBefore(void) const;

      /**  Get the duration of the generated waveform after the maximum time.
       */
      Interval    getTimeAfter(void) const;

      /**  Get the waveform amplitude
       */
      double      getAmplitude(void) const;

      /**  Get a list of the generated parameters.
       */
      const ParamList& getParamList(void) const;
      
    private:
      std::string mName;
      std::string mComment;
      std::string mInputs;
      Time        mGPSMax;
      Interval    mTimeBefore;
      Interval    mTimeAfter;
      double      mAmplitude;
      ParamList   mParamList;
    };

    typedef std::vector<SrcEvent> EventVect;
    typedef EventVect::iterator event_iter;
    typedef EventVect::const_iterator const_event_iter;
 
  public:
    /**  Static factory method produces data sources from a descriptor string
      *  of the form:
      *  \begin{verbatim}
           <data-source>(<param-1>=<value-1>,...,<param-n>=<value-n>)
         \end{verbatim}
      *  @memo Data source factory method.
      *  @param desc Data source descriptor string.
      *  @return Pointer to a new data source matching the description.
      */
    static DataSource* parse(const std::string& desc);

  public:
    /**  Default constructor
      */
    explicit DataSource(RepMode mode=kContinuous);

    /** Destroy and object
      */
    virtual ~DataSource(void);

    /** Make an exact copy of a data source
      */
    virtual DataSource* clone(void) const = 0;

    /**  Dump the data source status
     */
    virtual std::ostream& dump(std::ostream& out) const;


    /**  Dump the source definition string out to the specified output 
      *  stream.
      *  @memo print the source definition.
      *  @param out Output stream.
      *  @return reference to the output stream.
      */
    virtual std::ostream& dumpDefinition(std::ostream& out) const;

    /**  Test if daa source is continuous
     */
    bool isContinuous(void) const;

    /**  Generate events in a specified interval.
      */
    void generate(const Time& t0, Interval dt);

    /**  Get the data source type name.
     */
    virtual const char* getDataType(void) const = 0;

    /**  Get the debug level.
     */
    int    getDebug(void) const;

    /**  Get the length of an event.
     */
    virtual Interval getDuration(void) const;

    /**  Get the earliest time not generated.
      *  @memo Get the end of generated data.
      */
    Time getLatest(void) const;

    /**  Get the name of the data source.
     */
    const char* getName(void) const;

    /**  Get a numeric value of the named parameter.
     */
    virtual double      getNumeric(const std::string& name) const;

    /**  Get a string vaue of the named parameter.
     */
    virtual const char* getParameter(const std::string& name) const;

    /**  Get the average event rate.
     */
    double getRate(void) const;

    /**  Get the sample rate.
     */
    Interval getSample(void) const;

    /**  Get the offset from the generated event time to the start of the 
      *  event waveform. A positive offset corresponds to a signal that 
      *  starts after the generated time.
      */
    virtual Interval getStartOffset(void) const;

    /**  Get the start of the already generated data.
      */
    Time   getStartTime(void) const;

    /**  Get the trigger time fr the current event.
     */
    Time   getTrigTime(void) const;

    /**  Get the generated time series in the interval from t0 to t0+dT.
     */
    TSeries getTimeSeries(const Time& t0, Interval dT) const;

    /**  Get the generated data time series.
     */
    const TSeries& getTimeSeries(void) const;

    /**  Get the data source name.
     */
    virtual std::string getSourceName(void) const;

    /**  Print statistics for the data source
      */
    virtual std::ostream& print_stats(std::ostream& out) const;

    /**  Chose the time and parameters for the next data segment.
     */
    virtual void randomize(void);

    /**  Release generated data before the specified time.
      *  @param t0 Time before which data will be released.
      */
    void release(const Time& t0);

    /**  Set the debug level.
      */
    void setDebug(int lvl);

    /**  Set the data source name.
      */
    void setName(const std::string& name);

    /**  Set the specified parameter name to a string value.
      */
    virtual void setParameter(const std::string& name, const std::string& val);

    /**  Set the specified parameter to a numeric value.
      */
    virtual void setParameter(const std::string& name, double value);

    /** Set the digitization sample rate.
      */
    void setSample(Interval dT);

    /**  Set the average event rate.
     */
    void setTriggerRate(double rate);

    /**  Set the time for the next/only event.
      */
    void setTriggerTime(const Time& t0);

    //----------------------------------  Control which data are saved.
    /**  Get the write flag value. If true, the generate data are to be 
      *  written to an output frame.
      */
    bool getWriteFlag(void) const;    

    /**  Get the save flag value. If true, the event parameters are to be 
      *  saved for each event.
      */
    bool getSaveFlag(void) const;    

    /**  Set the write flags value.
      */
    void setWriteFlag(bool yorn=true);

    /** Set the save flag value
      */
    void setSaveFlag(bool yorn=true);

    //----------------------------------  Get event data
    /**  Get the number of events stored for this input source. If the save
      *  flag is not set, this number will always be zero.
      */
    unsigned long getNEvent(void) const;

    /**  Return a constant reference to the first event in the list of
      *  generated events.
      */
    const SrcEvent& refEvent(void) const;

    /**  Return a constant reference to the specified (ith) event in the
      *  list of generated events..
      */
    const SrcEvent& refEvent(unsigned long i) const;

    /**  Erase the specified number of events from the vector.
      *  @memo erase events
      *  @param N Number of events to erase.
     */
    void eraseEvent(unsigned long N);

  private:
    /**  Find the specified parameter name.
      */
    param_iter findParam(const std::string& name);
    const_param_iter findParam(const std::string& name) const;

    /**  Copy the generated source data from the specified time interval
      *  into the specified data vector.
      */
    virtual void getSeries(const Time& t0, Interval dT, int N, float* data)=0;

  private:
    //---------------------------------  Parameters
    RepMode  mRptMode;
    double   mRate;
    Interval mSample;
    int      mDebug;
    std::string mName;
    ParamMap mParam;
    enum Flags {
      flWrite=1,
      flSvEvents=2
    };
    long      mFlags;

    //---------------------------------  State
    Time     mNextTrig;
    Time     mCurrent;
    Time     mTrigStart;
    Time     mTrigEnd;
    Time     mLatest;
    TSeries  mSeries;
    long      mEventNo;
    EventVect mRecord;

    //----------------------------------  Statistics
    Interval  mSumTime;
  };  //class DataSource

  //====================================  Inline functions.
  inline bool
  DataSource::isContinuous(void) const {
      return mRptMode == kContinuous;
  }

  inline int 
  DataSource::getDebug(void) const {
      return mDebug;
  }

  inline Interval
  DataSource::getDuration(void) const {
      return 1.0;
  }

  inline Time 
  DataSource::getLatest(void) const {
      return mLatest;
  }

  inline const char*
  DataSource::getName(void) const {
      return mName.c_str();
  }

  inline Time
  DataSource::getTrigTime(void) const {
      return mNextTrig;
  }

  inline Interval
  DataSource::getSample(void) const {
      return mSample;
  }

  inline double
  DataSource::getRate(void) const {
      return mRate;
  }

  inline Interval
  DataSource::getStartOffset(void) const {
      return getDuration()*(-0.5);
  }

  inline Time
  DataSource::getStartTime(void) const {
      return mSeries.getStartTime();
  }

  inline bool 
  DataSource::getSaveFlag(void) const {
      return (mFlags & flSvEvents);
  }

  inline bool 
  DataSource::getWriteFlag(void) const {
      return (mFlags & flWrite);
  }

  inline DataSource::param_iter 
  DataSource::findParam(const std::string& name) {
      return mParam.find(name);
  }

  inline DataSource::const_param_iter 
  DataSource::findParam(const std::string& name) const {
      return mParam.find(name);
  }


  //----------------------------------  Get event data
  inline unsigned long 
  DataSource::getNEvent() const {
      return mRecord.size();
  }

  inline const DataSource::SrcEvent& 
  DataSource::refEvent(unsigned long i) const {
      return mRecord[i];
  }

  inline void 
  DataSource::eraseEvent(unsigned long N) {
      if (N) mRecord.erase(mRecord.begin(), mRecord.begin()+N);
  }

  inline const char* 
  DataSource::SrcEvent::getComment(void) const {
      return mComment.c_str();
  }

  inline const char* 
  DataSource::SrcEvent::getInputs(void)  const {
      return mInputs.c_str();
  }

  inline const char* 
  DataSource::SrcEvent::getName(void)    const {
      return mName.c_str();
  }

  inline const Time& 
  DataSource::SrcEvent::getGPSMax(void)  const {
      return mGPSMax;
  }

  inline Interval
  DataSource::SrcEvent::getTimeBefore(void) const {
      return mTimeBefore;
  }

  inline Interval
  DataSource::SrcEvent::getTimeAfter(void) const {
      return mTimeAfter;
  }

  inline  double
  DataSource::SrcEvent::getAmplitude(void) const {
      return mAmplitude;
  }

  inline const DataSource::SrcEvent::ParamList& 
  DataSource::SrcEvent::getParamList(void) const {
      return mParamList;
  }

} // namespace generator
#endif // GENERATOR_DATA_SOURCE_HH
