#include "Listen.hh"
#include "FilterDesign.hh"
#include "IIRFilter.hh"

#ifndef __CINT__
#include "AudioPanel.hh"
#include "TSeries.hh"
#include "TSystem.h"
#include "TROOT.h"
#include "TApplication.h"

EXEROOT(Listen)
#endif

using namespace std;

Listen::Listen(int argc, const char *argv[]) 
  : DatEnv(argc,argv), mPanel(0), mVolume(0.5), mMode(0), 
    mMute(false), mGain(0), mHigh(0), mLow(0), mFilter(0) 
{
    //----------------------------------  Parse command line
    bool syntax = false;
    if (argc < 2) {
        cerr << "Channel name not specified" << endl; 
	syntax = true;
    }

    for (int i=1 ; i<argc-1 && !syntax ; i++) {
        string argi = argv[i];
	if (argi == "-gain") {
	    mGain = strtod(argv[++i], 0);
	} else if (argi == "-high") {
	    mHigh = strtod(argv[++i], 0);
	} else if (argi == "-low") {
	    mLow = strtod(argv[++i], 0);
	} else if (argi == "-filter") {
	    mFDefine = argv[++i];
	} else if (isDatEnvArg(argv[i])) {
	    i++;
	} else {
	    cerr << "Argument: " << argv[i] << " not recognized." << endl;
	    syntax = true;
	}
    }
    if (syntax) {
        cerr << "Command syntax:" << endl;
	cerr << "Listen [<dmt-parameters>] [-gain <gain>] " 
	     << "[-filter <design-string>] <Channel>" << endl;
	finish();
	return;
    }
    mChannel=argv[argc-1];

    //----------------------------------  Build a control panel.
    mPanel = new AudioPanel;
    mPanel->setVolume(mVolume);
    mPanel->setMode(mMode);
    mPanel->setMute(mMute);
    gSystem->ProcessEvents();

    //----------------------------------  Connect to the Audio output channel
    try {
        mPort.open(0);
	if (Debug()) mPort.dumpStatus(cout);
	mPort.setVolume(mVolume);
	mPort.setPorts(mMode);
	mPort.mute(mMute);
	if (Debug()) mPort.dumpStatus(cout);
    } catch (runtime_error& e) {
        cout << "Error setting up audio port" << e.what() << endl;
	finish();
	return;
    }

    //----------------------------------  Get the channel to be monitored.
    getDacc().addChannel(mChannel.c_str());
    getDacc().setStride(8.0);
}

Listen::~Listen(void) {
}

void 
Listen::ProcessData(void) {
    const TSeries* ts = getDacc().refData(mChannel.c_str());
    double rate = 1.0/double(ts->getTStep());
    if (!mFilter && (mHigh || mLow)) {
        int nPoles(0);
        dComplex poles[4], zeros[4];
	if (mLow) {
	    poles[0] = -mLow;
	    poles[1] = -mLow;
	    zeros[0] = 0.0;
	    zeros[1] = 0.0;
	    nPoles = 2;
	}
	if (mHigh) {
	    poles[nPoles+0] = -mHigh;
	    poles[nPoles+1] = -mHigh;
	    zeros[nPoles+0] = -0.5*rate;
	    zeros[nPoles+1] = zeros[nPoles];
	    nPoles += 2;
	}
	IIRFilter pz(nPoles, poles, nPoles, zeros, rate, 1.0);
	mFilter = pz.clone();
    } else if (!mFilter && !mFDefine.empty()) {
        FilterDesign fd(mFDefine.c_str(), rate);
	mFilter = fd.release();
    }
    TSeries tsf;
    if (mFilter) tsf = (*mFilter)(*ts);
    else         tsf = *ts;
    if (mGain == 0) {
        mGain = 8192.0/(tsf.getMaximum() - tsf.getMinimum());
    }
    tsf *= mGain;

    //----------------------------------  Set the volume if it has changed
    mPanel->setTime(tsf.getStartTime());
    gSystem->ProcessEvents();
    bool changed = false;
    double Volume = mPanel->getVolume();
    if (Volume != mVolume) {
	mVolume = Volume;
        mPort.setVolume(mVolume);
	changed = true;
    }

    //----------------------------------  Set the output mode if it's changed
    int Mode = mPanel->getMode();
    if (Mode != mMode) {
	mMode = Mode;
        mPort.setPorts(mMode);
	changed = true;
    }

    //----------------------------------  Set the muting if it has changed
    bool Mute = mPanel->getMute();
    if (Mute != mMute) {
	mMute = Mute;
        mPort.mute(mMute);
	changed = true;
    }
    if (changed && Debug()) mPort.dumpStatus(cout);

    //----------------------------------  Spew a frame to the speakers
    int nSample = tsf.getNSample();
    short* iSample = new short[nSample];
    tsf.getData(nSample, iSample);
    if (!(getDacc().isOnline())) mPort.drain();
    mPort.reSample(double(1.0)/tsf.getTStep(), iSample, nSample);
    delete iSample;
}

void 
Listen::Attention(void) {
}
