/* -*- mode: c++; c-basic-offset: 4; -*- */
#define PIDCVSHDR "$Header: /ldcg_server/common/repository_gds/gds/Monitors/kleineWelle/kleineWelleM.cc,v 1.8 2009/06/03 18:00:59 lindy Exp $"
#define PIDTITLE  "kleineWelle"
#include "ProcIdent.hh"
#include "mywv2lib.hh"
#include "kleineWelleM.hh"
#include "LPEFilter.hh"
#include "DVector.hh"
#include "TrigBase.hh"

#include <iostream>
#include <fstream>
#include <iomanip>
#include <cstdlib>
#include <unistd.h>
#include <sstream>
#include <cmath>
#include <iterator>
#include <sys/stat.h>

#define KWVERSION "December 17, 2008"

EXECDAT(kleineWelle)

using namespace std;

//======================================  Channel data constructor
kwChan::kwChan(const std::string& name)
    : channel(name),              // name of the channel
      lowFrequency(64),           // low frequency bound of analysis
      highFrequency(1024),        // high frequency bound of analysis
      significance(10.0),         // significance threshold for trigger output
      threshold(3.0),             // threshold on sigmas for black pixels
      maxDist(2),                 // distance on time-scale plane for clustering
      decimateFactor(-1),         // times to decimate timeseries (-1 for auto)
      transientDuration(4.0),     // minimum time to cut for filter transients
      fs(0),                      // sample frequency of channel (0 for auto)
      wavePtr(NULL),
      lpefilter(NULL)
{
}

kwChan::~kwChan(void) {
    if(wavePtr) {
	delete wavePtr;
	wavePtr   = NULL;
    }
    if(lpefilter) {
	delete lpefilter;
	lpefilter = NULL;
    }
}

// Channel operator
kwChan& kwChan::operator=(const kwChan& x) {
    fs             = x.fs;
    channel        = x.channel;
    significance   = x.significance;
    threshold      = x.threshold;
    size           = x.size;
    lpefLength     = x.lpefLength;
    maxDist        = x.maxDist;
    lowFrequency   = x.lowFrequency;
    highFrequency  = x.highFrequency;
    decimateFactor = x.decimateFactor;
    transientDuration = x.transientDuration; // not used
    if (x.wavePtr) wavePtr = new Haar_Wavelet(size);
    else           wavePtr = NULL;
    prefilter      = x.prefilter; // auto-pipe clones filter.
    if (x.lpefilter) lpefilter  = x.lpefilter->clone();
    else             lpefilter  = NULL;
    return *this;
}

// set up sample-frequency derived data
void kwChan::setup(const TSeries* ts) {
    size = ts->getNSample();
    fs = int(size / ts->getInterval() + 1e-6);
    FilterDesign fd(fs);
    if(highFrequency > fs / 2)  // cap to nyquist
	highFrequency = fs / 2;
    lowScale  = (int)ceil(log2(fs / highFrequency));
    highScale = (int)ceil(log2(fs / lowFrequency));
    /// decimator is part of prefilter
    if(decimateFactor == -1)
	decimateFactor = lowScale - 1;  // leave headroom for linear filter
    fs = (fs >> decimateFactor);
    // need to adjust wavelet parameters
    size = (size >> decimateFactor);
    lowScale -= decimateFactor;
    highScale -= decimateFactor;
    // prefilter
    fd.decimateBy2(decimateFactor, 3); // type 3 has most ringing, but least timeshift
    fd.butter(kHighPass, 6, lowFrequency, 0);
    prefilter.set(fd.release());
    // construct LEPF
    lpefLength = min(int(double(transientDuration) * fs), 
		     int(16*highFrequency/lowFrequency));
    if (lpefilter) delete lpefilter;
    lpefilter = new LPEFilter(lpefLength, 0, size-1);
    if (wavePtr) delete wavePtr;
    wavePtr   = new Haar_Wavelet(size);          // initialize wavelet class
}

// grab timeseries and train/apply filters, put events in events vector
event_vect 
kwChan::process(MultiDacc& dacc) {
    TSeries* tsIn = dacc.refData(channel.c_str());
    tsIn->Convert(DVector::t_float);
    Time t0 = tsIn->getStartTime();
    if(fs == 0) setup(tsIn);         // first time data is looked at, do setup
    TSeries tsHpf = prefilter(*tsIn); //autopipe::operator()
    lpefilter->train(tsHpf);
    lpefilter->apply(tsHpf, tsLpef);
    tsLpef.getData(size, wavePtr->getRawArray()); // put data into wavelet class
    wavePtr->transform(1);                     // wavelet transform
    const double* wavearray = wavePtr->getWaveArray();
    element_vect elements;
    // build vector of black pixels
    for(int scale = lowScale; scale <= highScale; scale++) {
	double dt = double(1 << scale) / double(fs);  // duration of one pixel
	int lsize = size / (1 << scale);  // index where array for scale begins
	double lsum = 0;
	for(int i = lsize; i < 2*lsize; i++) // calculate sigma
	    lsum += wavearray[i]*wavearray[i];
	double sigma = sqrt(lsum / lsize);
	for(int i = lsize; i < 2*lsize; i++) {
	    double amplitude = wavearray[i];
	    double normalized = amplitude / sigma;
	    if(abs(normalized) > threshold) {
		double time = dt * double(i-lsize);
		elements.push_back(Element(amplitude*amplitude, 
					   normalized*normalized,
					   i, scale, time, 
					   time + dt, dt, 0));
	    }
	}
    }

    // assign cluster ID to black pixels
    int freeID = 0;
    int N = elements.size();
    if (N > 0) {
	vector<int> pvect(N, 0);
	for (int i=0; i<N; ++i)
	    pvect[i] = i;
	int ifree = 0;
	for (int i=0; i<N; ++i) {
	    if (i == ifree) {
		ifree++;
		elements[pvect[i]].clusterID = freeID++;
	    }
	    const Element& testEl(elements[pvect[i]]);
	    for (int j=ifree; j<N; ++j) {
		int pj = pvect[j]; 
		if (testEl.dist(elements[pj]) <= maxDist) {
		    elements[pj].clusterID = testEl.clusterID;
		    if (j != ifree) {
			pvect[j]     = pvect[ifree];
			pvect[ifree] = pj;
		    }
		    ifree++;
		}
	    }
	}
    }

    // sum clusters by ID
    std::stringstream out;
    out << channel << "_" << lowFrequency << "_" << highFrequency;
    uniqueID = out.str();
    uniqueID.replace(2, 1, "_");
    event_vect events(freeID, Event(uniqueID));
    for(element_iter i=elements.begin(); i < elements.end(); i++) {
	events[i->clusterID] += *i;
    }

    // estimate cluster parameters
    int j=0;
    for(int i = 0; i < freeID; i++) {
	events[i].average(t0, fs);
	if(events[i].significance >= significance) {
	    if (j != i) events[j] = events[i];
	    j++;
	}
    }

    // return sorted events
    events.erase(events.begin()+j, events.end());
    sort(events.begin(), events.end());
    return(events);
}

kleineWelle::kleineWelle(int argc, const char *argv[])
    : MultiStream(argc, argv),
      mTrigClient(trig::TrigWriter::kWriter),
      stride(16.0),
      basename("KW"),
      lasttime(0)
{
    bool syntax = false;        // flag for syntax errror

    // clear process XML output
    mTrigClient.setTableFile("/dev/null");
    mTrigClient.flush(Time(0, 0), Time(0, 0));

    // parse command syntax
    if(argc == 1) {
	syntax = true;
    } else {
	optionsFile.open(argv[1], ios::in);
        if(!(optionsFile.is_open())) {
            cerr << "cannot open options file: " << argv[1] << endl;
            finish();
            return;
        } else if (Debug()) {
            cerr << "reading options file: " << argv[1] << endl;
        }
    }
    if(syntax) {
        cerr << "kleineWelle version " << KWVERSION << endl;
        cerr << "Command syntax:" << endl;
        cerr << argv[0] << " [<options file>] [<MultiStream-args>]" << endl;
        cerr << "please see sample options file 'optM' for example" << endl;
        finish();
        return;
    }
    // read in options from file
    string segment_file_name;
    kwChan default_chan = kwChan("default");
    kwChan temp = kwChan("default");
    /* not currently used
    kwVeto default_veto = kwVeto("");
    kwVeto tempv = kwVeto("");
    */
    while(!optionsFile.eof()) {
        string option, word;
        getline(optionsFile, option);
        if (Debug()) cerr << "Option: " << option << endl;
        istringstream optionsStream(option);
        optionsStream >> word;
        if(word.length() == 0 || word[0] == '#') {
            if (Debug()) cerr << "Option skipped" << endl;
            continue;
        }
        else if(word == "stride") {         // DMT analysis stride
            double epoch;
            optionsStream >> epoch;
            stride = Interval(epoch);
        }
	else if(word == "basename")         // prefix for trigger files
	    optionsStream >> basename;
        else if(word == "threshold")        // default threshold on coefficient amplitude
            optionsStream >> default_chan.threshold;
        else if(word == "significance")     // default threshold on cluster significance
            optionsStream >> default_chan.significance;
        else if(word == "maxDist")          // default distance for clustering
            optionsStream >> default_chan.maxDist;
        else if(word == "decimateFactor")   // times to decimate data (-1 = auto, 0 = none)
            optionsStream >> default_chan.decimateFactor;
        else if(word == "transientDuration") { // length of filter transient to remove
            double epoch;
            optionsStream >> epoch;
            default_chan.transientDuration = Interval(epoch);
        }
	else if(word == "vetoseg") {
        /* not currently used
	    tempv = default_veto;
	    optionsStream >> tempv.conffile;
	    optionsStream >> tempv.nfactor;
        */
	}
        else if(word == "channel") {        // KW analysis channel
            temp = default_chan;            // initialize channel with default values 
	    optionsStream >> temp.channel;
	    optionsStream >> temp.lowFrequency;
	    optionsStream >> temp.highFrequency;
	    optionsStream >> temp.significance;
            mChanList.push_back(temp);
	    if (Debug()) cerr << "Configure channel: " << temp.channel << endl;
        }
        else
            cerr << "bad option: " << word << endl;
    }
    optionsFile.close();

    // Check at least one channel
    if (mChanList.empty()) {
        cerr << "No channels were configured!" << endl;
        finish(); 
        return;
    } else {
	for (channel_iter i=mChanList.begin(); i != mChanList.end(); ++i)
	    getDacc().addChannel(i->channel.c_str(), 0);
    }


    // Set up data stream.
    getDacc().setStride(stride);
    setStrideAlignment(long(stride), 0.0);
}

kleineWelle::~kleineWelle()
{
    cout << "# finished." << endl;
}

void kleineWelle::ProcessData(void)
{
    Time t0 = getDacc().getFillTime();
    double rem = fmod(t0.totalS(), stride.GetSecs());
    cout << "processing at " << t0 << " offset " << t0.totalS() 
	 << " " << stride.GetSecs() << " " << rem << endl;

    // do we need tolerance here? (not as long as the stride is an int).
    if(rem == 0) {
	std::stringstream path;
	path << basename << "-" << t0.getS()/100000;
	std::string dir = path.str();
	path << "/" << basename << "-" << t0.getS() << "-" 
	     << stride.GetSecs() << ".trg";
	std::string file = path.str();

	ofstream outfile;
	outfile.open(file.c_str());
	if(!outfile) {
	    mkdir(dir.c_str(), 0755);
	    outfile.open(file.c_str());
	    if(!outfile) {
		cerr << "cannot create file: " << file;
		finish();
	    }
	}

	// check state vector
	if(!lasttime || t0 == lasttime) {
	    for(channel_iter i=mChanList.begin(); i != mChanList.end(); ++i) {
		event_vect events = i->process(getDacc());
		copy(events.begin(), events.end(), ostream_iterator<Event>(outfile));
            // fill trigger table
            for(event_iter i=events.begin(); i != events.end(); ++i) {
               Time::ulong_t sec  = (Time::ulong_t)i->startTime;
               Time::ulong_t nsec = (Time::ulong_t)(1e9 * (i->startTime - sec));
               trig::TrigBase trigger(i->channelName.c_str(), "", Time(sec, nsec), Interval(i->endTime - i->startTime),
                                      sqrt(i->normEnergy - i->npts), i->significance, i->frequency, "");
               trigger.setAvgOffset(i->centralTime - i->startTime);
               mTrigClient.sendTrigger(trigger);
            }
	}
        // output XML version
        file.replace(file.size() - 3, 3, "xml");
        mTrigClient.setTableFile(file.c_str());
        mTrigClient.flush(t0, t0+stride);
	} else {
	    cout << "suppressing trigger output at break in data, expected " 
		 << t0 << " found " << lasttime+stride << endl;
	    for(channel_iter i=mChanList.begin(); i != mChanList.end(); ++i) {
		event_vect events = i->process(getDacc());
	    }
	}
	lasttime = getDacc().getCurrentTime(); 
	outfile.close();
    } else {
	Interval wait = stride - Interval(rem);
	Time t2 = t0 + wait;
	cout << "aligning with stride skipping " << wait 
	     << " seconds to " << t2 << endl;
	getDacc().seek(t2);
	lasttime = Time(0);
    }
}

//======================================  Event null constructor
Event::Event(const std::string& name)
    : startTime(0), endTime(0), centralTime(0), frequency(0),
      energy(0), normEnergy(0), npts(0), significance(0),
      channelName(name)
{}

//======================================  Add an element
Event&
Event::operator += (const Element& el) {
    if (npts++) {
        startTime = min(startTime, el.startTime);
        endTime   = max(endTime, el.endTime);
    } else {
        startTime = el.startTime;
        endTime   = el.endTime;
    }
    double norm = el.normEnergy;
    centralTime += norm * (el.startTime + el.endTime)/2.;
    frequency += el.scale * norm;
    energy += el.energy;
    normEnergy += norm;
    return *this;
}

//======================================  Calculated event parameters
void
Event::average(const Time& toff, double fs) {
    if (!npts) return;
    //1/centralTime = centralTime / normEnergy;
    centralTime = toff.totalS() + centralTime / normEnergy;
    startTime += toff.totalS();
    endTime   += toff.totalS();
    //1/----------------------------------------
    significance = logSignificance(normEnergy/2.0, npts/2.0, .01);
    frequency = fs / pow(2.0, frequency / normEnergy);
}

//======================================  Event printout operator.
std::ostream&
operator<<(std::ostream& output, const Event& a) {
    char line[128];
    sprintf(line, "%16.6f  %16.6f  %16.6f  %4.0f  %9.3e  %6.2f  %3d  %6.2f  %s",
            a.startTime, a.endTime, a.centralTime, a.frequency, a.energy,
            a.normEnergy, a.npts, a.significance, a.channelName.c_str());
    output << line << endl;
    return output;
}

// negative natural logarithm of significance
double
logSignificance(double normalizedEnergy, double timeFrequencyArea,
                double precision)
{

    /*
     * This function assumes time-frequency measurement of stationary white
     * Gaussian noise.  The resulting measurement (2*normalizedEnergy) are chi squared
     * distributed with 2A degrees of freedom, where A is the time-frequency
     * area of the measurement.  This function returns the negative natural
     * logarithm of the probability of achieving a less likely result.  It
     * uses a series expansion of the regularized upper incomplete gamma
     * function which is valid for the tail (highly significant) portion of
     * the distribution.  See Ambramowitz and Stegun, Handbook of Mathematical
     * Functions, section 6.1, page 264.
     */

    // maximum number of terms in series expansion
    int maximumTerms = 15;

    // initialize series expansion
    double sum = 1;
    double numerator = 1;
    double denominator = 1;
    double ratio;
    double term;

    // compute and sum series terms
    for (term = 1; term <= maximumTerms; term++) {
	numerator *= (timeFrequencyArea - term) * (timeFrequencyArea > term);
	denominator *= normalizedEnergy;
	ratio = numerator / denominator;
	sum += ratio;
	if (ratio < precision * sum) {
	    break;
	}
    }

    // give error if maximum number of terms was exceeded
    if (term > maximumTerms) {
	std::cerr << "logSignificance: maximum number of terms exceeded"
		  << std::endl;
    }

    // negative natural logarithm of significance
    double logSignificance = normalizedEnergy -
	(timeFrequencyArea - 1) * log(normalizedEnergy) -
	log(sum) + lgamma(timeFrequencyArea);

    // return to calling function
    return logSignificance;

}
